"""
This module contains unit tests for the WWZ.
"""

import unittest


class TestQuantum(unittest.TestCase):
    def test_q(self):
        """
        I don't do much yet...
        """
        self.assertTrue(True)
